import os
import mock
import shutil
import tempfile
import unittest
import yaml

from reactive import node_exporter
from charmhelpers.core import unitdata
from charmhelpers.core.templating import render
from charms.reactive import bus

os.environ['JUJU_UNIT_NAME'] = 'node-exporter'
os.environ['CHARM_DIR'] = '..'


class SimpleConfigMock(dict):
    def __init__(self, *arg, **kw):
        super(SimpleConfigMock, self).__init__(*arg, **kw)
        self._changed_dict = {}
        self._changed_default = True

    def changed(self, key):
        return self._changed_dict.get(key, self._changed_default)

    def set_changed(self, changed_dict):
        self._changed_dict.update(changed_dict)


@mock.patch('os.chown')
@mock.patch('os.fchown')
@mock.patch('reactive.node_exporter.hookenv.status_set')
@mock.patch('reactive.node_exporter.host.service_restart')
@mock.patch('reactive.node_exporter.host.service_running')
@mock.patch('reactive.node_exporter.data_changed')
@mock.patch('reactive.node_exporter.hookenv.unit_get')
@mock.patch('reactive.node_exporter.hookenv.config')
class TestPrometheusContext(unittest.TestCase):
    def setUp(self):
        super(TestPrometheusContext, self).setUp()
        self._init_tempdir_and_filenames()
        self._init_load_config_yml_defaults()
        self._init_override_render_for_nonroot()

    def _init_tempdir_and_filenames(self):
        self.dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.dir)
        self.maxDiff = None
        self.cfg_def = os.path.join(self.dir, 'test_etc_default_node_exporter')
        node_exporter.PUSHGATEWAY_DEF = self.cfg_def
        # ugly hack, to avoid carrying global unitdata state across tests
        os.environ['UNIT_STATE_DB'] = os.path.join(self.dir, '.unit-state.db')
        unitdata._KV = None

    def _init_load_config_yml_defaults(self):
        # create def_config with parsed config.yaml defaults
        with open("../config.yaml") as fh:
            config_yaml = yaml.safe_load(fh)
            self.def_config = SimpleConfigMock({k: v["default"] for k, v in
                                                config_yaml["options"].items()
                                                if v.get("default")})

    def _init_override_render_for_nonroot(self):
        # need to override render() for 644 perms (ie allow overriding target
        # when running as non-root)
        def render_and_chmod(*args, **kwargs):
            render(*args, **kwargs)
            os.chmod(kwargs['target'], 0o664)
        self.mock_render = mock.patch('reactive.node_exporter.render',
                                      side_effect=render_and_chmod)
        self.mock_render.start()
        self.addCleanup(self.mock_render.stop)

    def test_other_configs(self,
                           mock_hookenv_config,
                           mock_hookenv_unit_get,
                           mock_data_changed,
                           mock_service_running,
                           mock_service_restarted,
                           mock_status_set,
                           *args):
        config = self.def_config
        config.update({
            'port': '9876',
            'textfile_directory': os.path.join(self.dir, 'foo/bar')
        })
        exp_args = [
            '-web.listen.address :{}'.format(
                config['port']),
            '-collector.textfile.directory {}'.format(
                config['textfile_directory']),
        ]
        mock_hookenv_config.return_value = config
        node_exporter.write_node_exporter_config_def()
        # Verify etc/default/prometheus-node-exporter has
        # -web.external-url added
        with open(self.cfg_def) as fh:
            contents = fh.read()
            for args in exp_args:
                self.assertRegexpMatches(
                    contents, 'ARGS=.*{}.*'.format(args))
        # Verify textfile_directory created
        self.assertTrue(os.path.isdir(config['textfile_directory']))

    @mock.patch('reactive.node_exporter.set_state')
    def test_install_packages_conditionally_called(self,
                                                   mock_set_state,
                                                   mock_hookenv_config,
                                                   *args):
        config = self.def_config
        mock_hookenv_config.return_value = config
        # No configs changed, expect no node_exporter-do-install
        config.set_changed({'install_sources': False, 'install_keys': False})
        node_exporter.check_reconfig_node_exporter()
        self.assertFalse(mock.call('node-exporter.do-install')
                         in mock_set_state.call_args_list)
        # Mock some configs changed, expect node_exporter-do-install
        mock_set_state.reset_mock()
        config.set_changed({'install_sources': True, 'install_keys': False})
        node_exporter.check_reconfig_node_exporter()
        self.assertTrue(mock.call('node-exporter.do-install')
                        in mock_set_state.call_args_list)

    def test_configure_node_exporter_relation(self,
                                              mock_hookenv_config,
                                              *args):
        mock_hookenv_config.return_value = self.def_config
        rel_mock = mock.Mock()
        # Verify configure_node_exporter_relation calls configure(port)
        node_exporter.configure_node_exporter_relation(rel_mock)
        self.assertListEqual(rel_mock.method_calls,
                             [mock.call.configure(self.def_config['port'])])

    def test_runtime_args(self,
                          *args):
        node_exporter.runtime_args('-foo', '1')
        node_exporter.runtime_args('-bar', '2')
        self.assertListEqual(node_exporter.runtime_args(), ['-bar 2', '-foo 1'])
        node_exporter.runtime_args('-bar', None)
        self.assertListEqual(node_exporter.runtime_args(), ['-foo 1'])

    @mock.patch('charms.reactive.bus.SourceFileLoader')
    @mock.patch('reactive.node_exporter.fetch')
    @mock.patch('reactive.node_exporter.install_packages')
    # Simulate a full run, from zero to final state,
    # verify we end calling service_restarted(...)
    def test_started_after_full_run(self,
                                    mock_install_packages,
                                    mock_fetch,
                                    mock_sourcefileloader,
                                    mock_hookenv_config,
                                    mock_hookenv_unit_get,
                                    mock_data_changed,
                                    mock_service_running,
                                    mock_service_restarted,
                                    *args):

        # Feed default config
        mock_hookenv_config.return_value = self.def_config
        # Force data_changed() always true
        mock_data_changed.return_value = True

        # Simulate a full reactive run - will consume states until settled:
        bus.dispatch()

        # Verify: /etc/default/prometheus-node-exporter present,
        # and calls to: install_packages(), service_restarted(...)
        self.assertTrue(os.path.exists(self.cfg_def))
        mock_install_packages.called_once_with()
        mock_service_running.called_once_with(node_exporter.SVCNAME)
        mock_service_restarted.called_once_with(node_exporter.SVCNAME)
